#!/usr/bin/env python3

import sys
import os
from video_fft.video_fft_calculator import VideoFftCalculator
import pytest

sys.path.insert(0, os.path.abspath(os.path.join(os.path.dirname(__file__), "..")))

import video_fft.__main__ as video_fft  # noqa E402

this_dir = os.path.abspath(os.path.dirname(__file__))
input_file = os.path.join(this_dir, "test.mp4")

ground_truth = {
    "average_fft": 14310,
    "max_fft": 14313,
    "min_fft": 14307,
    "median_fft": 14310,
    "pct_05": 14308,
    "pct_95": 14313,
}


def test_images():
    fft = VideoFftCalculator(input_file, all_frames=True, mean=True)
    fft.calc_fft()

    files = [f"test_frame-{i}.png" for i in range(5)]
    files.append("test-mean.png")

    for f in files:
        fpath = os.path.join(this_dir, f)
        assert os.path.isfile(fpath)
        try:
            os.unlink(fpath)
        except Exception as e:
            print(e)
            pass

def test_stats():
    fft = VideoFftCalculator(input_file)
    fft.calc_fft()
    ret = fft.get_stats()

    del ret["input_file"]

    assert ret == ground_truth



import numpy as np
from matplotlib import pyplot as plt
import json
import os
from tqdm import tqdm
from typing import Generator
import av
import sys


class VideoFftCalculator:
    def __init__(
        self,
        input_file,
        num_frames=None,
        output=None,
        first_frame=False,
        all_frames=False,
        mean=False,
        scale=1,
        output_format="json",
        quiet=False,
    ):
        self.input_file = str(input_file)
        self.output = (
            str(output) if output is not None else os.path.dirname(self.input_file)
        )
        self.num_frames = int(num_frames) if num_frames is not None else None
        self.first_frame = bool(first_frame)
        self.all_frames = bool(all_frames)
        self.mean = bool(mean)
        self.scale = float(scale)
        self.output_format = str(output_format)
        self.quiet = bool(quiet)

        self.frame_height = None
        self.frame_width = None
        self.fig_prefix = os.path.join(
            self.output, os.path.splitext(os.path.basename(self.input_file))[0]
        )

        self.data = {}
        self.last_magnitude_spectrum = None  # for later
        self.avg_magnitude_spectrum = None  # for later

        if self.output_format not in ["json", "csv"]:
            raise RuntimeError("Wrong output format, must be 'json' or 'csv'")

    @staticmethod
    def radial_profile(data, center):
        """
        Source: https://stackoverflow.com/a/21242776/435093
        """
        y, x = np.indices((data.shape))
        r = np.sqrt((x - center[0]) ** 2 + (y - center[1]) ** 2)
        r = r.astype(int)

        tbin = np.bincount(r.ravel(), data.ravel())
        nr = np.bincount(r.ravel())
        radialprofile = tbin / nr
        return radialprofile

    def get_stats(self):
        if not self.data:
            raise RuntimeError(
                "Data has not been generated yet, call calc_fft() first!"
            )

        return self.data

    def print_stats(self) -> None:
        if not self.data:
            raise RuntimeError(
                "Data has not been generated yet, call calc_fft() first!"
            )

        if self.output_format == "json":
            print(json.dumps(self.data, indent=4))
        elif self.output_format == "csv":
            import csv

            writer = csv.writer(sys.stdout)
            writer.writerow(self.data.keys())
            writer.writerow(self.data.values())
        else:
            # shouldn't happen
            pass

    def get_num_frames(self) -> int:
        container = av.open(self.input_file)
        num_frames = container.streams.video[0].frames
        return num_frames

    def read_input_file(self) -> Generator[np.ndarray, None, None]:
        container = av.open(self.input_file)

        if not len(container.streams.video):
            raise RuntimeError("No video streams found!")

        for frame in container.decode(video=0):
            frame_data = (
                frame.to_ndarray(format="gray")
                .reshape(frame.height, frame.width)
                .astype("float32")
            )
            yield frame_data

    def plot_magnitude_spectrum(self, what: str, frame_index=None):
        if what not in ["last", "mean"]:
            raise RuntimeError("'what' must be 'last' or 'mean'")

        data = (
            self.last_magnitude_spectrum
            if what == "last"
            else self.avg_magnitude_spectrum
        )

        if data is None:
            raise RuntimeError("No data to plot!")

        plt.figure(
            figsize=(
                (self.frame_width / 100) * self.scale,
                (self.frame_height / 100) * self.scale,
            ),
            dpi=72,
        )
        plt.imshow(data, cmap="gray")

        if what == "last":
            if frame_index is None:
                raise RuntimeError("Pass a frame index!")
            file_path = f"{self.fig_prefix}_frame-{frame_index}.png"
        else:
            file_path = f"{self.fig_prefix}-mean.png"

        plt.savefig(file_path, bbox_inches="tight")

        print(f"File written to {file_path}", file=sys.stderr)

    def calc_fft(self):
        # Generate empty list fft sum values
        fft_values = []

        if self.num_frames is not None:
            num_frames = min(self.get_num_frames(), self.num_frames)
        else:
            num_frames = self.get_num_frames()

        t = tqdm(total=num_frames, disable=self.quiet, file=sys.stderr)

        current_frame = 0
        for frame_data in self.read_input_file():
            fshift = np.fft.fftshift(np.fft.fft2(frame_data))
            self.last_magnitude_spectrum = 20 * np.log(np.abs(fshift))

            if self.avg_magnitude_spectrum is None:
                self.avg_magnitude_spectrum = self.last_magnitude_spectrum
            else:
                self.avg_magnitude_spectrum = np.mean(
                    np.array(
                        [self.avg_magnitude_spectrum, self.last_magnitude_spectrum]
                    ),
                    axis=0,
                )

            if not (self.frame_height or self.frame_width):
                self.frame_height, self.frame_width = self.last_magnitude_spectrum.shape

            if (self.first_frame and current_frame == 0) or self.all_frames:
                self.plot_magnitude_spectrum("last", current_frame)

            center = (self.frame_width / 2, self.frame_height / 2)

            # Calculate the azimuthally averaged 1D power spectrum
            psf_1d = VideoFftCalculator.radial_profile(
                self.last_magnitude_spectrum, center
            )
            # Get the sum of the high frequencies
            low_freq_ind = int((len(psf_1d) / 2))
            fft_values.append(sum(psf_1d[low_freq_ind:]))

            current_frame += 1
            t.update()
            if num_frames is not None and current_frame >= num_frames:
                break

        t.close()

        if self.mean:
            self.plot_magnitude_spectrum("mean")

        if not fft_values:
            raise RuntimeError("No FFT values were calculated!")

        self.data = {
            "input_file": os.path.abspath(self.input_file),
            "average_fft": int(np.mean(fft_values)),
            "max_fft": int(max(fft_values)),
            "min_fft": int(min(fft_values)),
            "median_fft": int(np.median(fft_values)),
            "pct_05": int(np.percentile(fft_values, 5, interpolation="midpoint")),
            "pct_95": int(np.percentile(fft_values, 95, interpolation="midpoint")),
        }



#!/usr/bin/env python3
from argparse import ArgumentParser, ArgumentDefaultsHelpFormatter

from .video_fft_calculator import VideoFftCalculator
from . import __version__ as version


def main():
    parser = ArgumentParser(
        formatter_class=ArgumentDefaultsHelpFormatter,
        description=f"video-fft v{version}",
    )
    parser.add_argument("input", help="Input video file")
    parser.add_argument(
        "-o",
        "--output",
        type=str,
        help="Output path for the images, default: same as input video file",
    )
    parser.add_argument(
        "-n", "--num-frames", type=int, help="Number of frames to calculate"
    )
    parser.add_argument(
        "-of",
        "--output-format",
        help="Select output format",
        choices=["json", "csv"],
        default="json",
    )
    parser.add_argument(
        "-r",
        "--first-frame",
        help="Render image for radial profile of the first frame",
        action="store_true",
    )
    parser.add_argument(
        "-a",
        "--all-frames",
        help="Render image for radial profile of all frames",
        action="store_true",
    )
    parser.add_argument(
        "-m",
        "--mean",
        help="Render image for mean/average radial profile of the entire sequence",
        action="store_true",
    )
    parser.add_argument(
        "-s",
        "--scale",
        help="Image scaling, adjust to increase/decrease rendered image size",
        default=1,
        type=float
    )
    parser.add_argument(
        "-q",
        "--quiet",
        help="Do not show progress bar",
        action="store_true",
    )
    cli_args = parser.parse_args()

    vid_fft = VideoFftCalculator(
        cli_args.input,
        num_frames=cli_args.num_frames,
        output=cli_args.output,
        first_frame=cli_args.first_frame,
        all_frames=cli_args.all_frames,
        mean=cli_args.mean,
        scale=cli_args.scale,
        output_format=cli_args.output_format,
        quiet=cli_args.quiet
    )
    vid_fft.calc_fft()
    vid_fft.print_stats()


if __name__ == "__main__":
    main()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
from setuptools import setup

# To use a consistent encoding
from codecs import open
from os import path

here = path.abspath(path.dirname(__file__))

# read version string
with open(path.join(here, "video_fft", "__init__.py")) as version_file:
    version = eval(version_file.read().split(" = ")[1].strip())

# Get the long description from the README file
with open(path.join(here, "README.md"), encoding="utf-8") as f:
    long_description = f.read()

# Get the history from the CHANGELOG file
with open(path.join(here, "CHANGELOG.md"), encoding="utf-8") as f:
    history = f.read()

setup(
    name="video_fft",
    version=version,
    description="Calculate the magnitude spectrum of a video sequence, via Fast Fourier Transform",
    long_description=long_description + "\n\n" + history,
    long_description_content_type="text/markdown",
    author="Werner Robitza",
    author_email="werner.robitza@gmail.com",
    url="https://github.com/slhck/video-fft",
    packages=["video_fft"],
    include_package_data=True,
    install_requires=[
        "av>=8.0.3",
        "tqdm>=4.60.0",
        "numpy>=1.20.2",
        "matplotlib>=3.4.2",
    ],
    license="MIT",
    zip_safe=False,
    keywords="ffmpeg, video",
    classifiers=[
        "Development Status :: 4 - Beta",
        "Intended Audience :: Developers",
        "Topic :: Multimedia :: Video",
        "License :: OSI Approved :: MIT License",
        "Natural Language :: English",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
    ],
    python_requires=">=3.10",
    entry_points={"console_scripts": ["video-fft = video_fft.__main__:main"]},
)
